%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (CHEM-E7225) ADVANCED PROCESS CONTROL D
%   #01 - Task 1 (Explicit integrators methods)
%
%   Complete this starting code (and auxiliary functions) to simulate 
%   the benchmark Three-Tank system described in Task 1.
%
%   Find and replace each '???' with the correct code needed to simulate
%   the model in ThreeTank.m .
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all; clear; clc

addpath('path/to/casadi')   % <~ Change to the appropriate path in your computer
import casadi.*

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECTION I - Define the initial value problem (IVP)

% Simulation parameters
Tf = ???;   
K  = ???;   
dT = ???;   

% Model parameters
S_T = ???;  S_V = ???; a_V = ???; % and so on...

px = ???;

% Initial state
x0 = ???;

% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define the input signal u(t) to drive the system

u1 = ???;       % <~~ Define the input signal u_1(t)
u2 = ???;       % <~~ Define the input signal u_2(t)
u = [u1; u2];


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECTION II - Explicit Euler
%  Numerical method to solve IVPs given xplus = f(t,x,u|p) and initial x(0),
%  over [0,Tf], by sequentially solving the formula
%
%       x_k+1 = x_k + f(k*dT,x_k,u_k|p)*dT
%             = F_EE(x_k,u_k)
%
%  for k = [0,1,...,Tf/dT] and x_0 = x(0), given a step-size dT > 0.

F_ee = @(x,u) eefnc(@ThreeTank, dT, x, u, px);

X_ee = nan(???,???);    % <~~ Fix the state-space dimension
X_ee(:,1) = x0;

for k = 1:K-1
  ???                   % <~~ Complete the simulation loop
end


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECTION III - Explicit Range-Kutta Order 4 (RK4)
%  Numerical method to solve IVPs given xplus = f(t,x,u|p) and initial x(0),
%  over [0,Tf], by sequentially solving the formulas
%
%       k1 = f(k*dT     , x_k        , u_k|p)
%       k2 = f(k*dT+dT/2, x_k+dT*k1/2, u_k|p)
%       k3 = f(k*dT+dT/2, x_k+dT*k2/2, u_k|p)
%       k4 = f(k*dT+dT  , x_k+dT*k3  , u_k|p)
%       x_k+1 = x_k + dT/6*(k1 + 2*k2 + 2*k3 + k4)
%             = F_RK4(x_k,u_k)  
%
%  for k = [0,1,...,Tf/dT] and x_0 = x(0), given a step-size dT.

F_rk4 = @(x,u) rk4fnc(@ThreeTank, dT, x, u, px);

X_rk4 = nan(???,???);       % <~~ Fix the state-space dimension
X_rk4(:,1) = x0;

for k = 1:K-1
  ???                       % <~~ Complete the simulation loop
end


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECTION IV - CasADi (using the Integrator function)
%  Define an integrator using an ODE symbolic expression 
%
%     x_k+1 = F_ode(x|p_ode), 
%
%  with the parameters p_ode = [u, p].

X = MX.sym( ??? );    % <~~ Define correctly the respective 
U = MX.sym( ??? );    % <~~ symbols with their correct dimensions
P = MX.sym( ??? );    % <~~ - 

ODE = ???;            % <~~ Define the ODE object for the integrator
F_casadi = integrator('F', 'rk', ODE, struct('number_of_finite_elements',1, 'tf',dT));

X_casadi = nan(???,???);    % <~~ Fix the state-space dimension
X_casadi(:,1) = x0;

for k = 1:K-1
  ???                       % <~~ Complete the simulation loop
end

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECTION V - Forward Simulation Map (using an linearization of f(.))
%  Analytic solution for a (affine)linear time-invariant (LTI) model
%
%     x_k = A^(k-1)x0 + \sum_(j=1)^(k-1) A^(k-j-1)(B u_j + c)
%         = F_sim(x0, u_1,...,u_{k-1}), 
%
%  with affine c = f(xp,up|px) and Jacobian matrices A = (dF/dx)|_(xp,up) 
%  and B = (dF/du)|_(xp,up) evaluated at point P = (xp,up) = (x0,u0).
%  (Obs.:  Here, F(.) denotes the analytic solution map of f(.))
%  (Obs.2: The formula differs from the slides because MATLAB's indexing starts at 1)

xp = x0; up = u(:,1);    % The initial conditions are used as the linearization point

ThreeTank_Lin = Linearize(@ThreeTank, dT, xp, xp, px);   
F_sim = @(x0,u) fsimfnc(ThreeTank_Lin, x0, u);   

%
X_lin = nan(???,???);    % <~~ Fix the state-space dimension

for k = 1:K
  ???                    % <~~ Complete the forward simulation function
end
  

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SECTION VI - Plotting all the simulation results
%
%  NOTE: You don't need to complete this part, but feel free to tweak 
%        the parameters to customize the plots to suite your style. 
%
X_data = {X_ee, X_rk4, X_casadi, X_lin};

t = (0:K-1)*dT;
ms = 4.0; lw = 1.0; fs = 14;
figure(1), clf; 
for i = 1:4
  for j = 1:2
    subplot(2,3,j+(j==2)), 
      stairs(t, u(j,:), 'k-')
      ylabel("$u_"+j+"(t)$ [ml/s]",'Interpreter','Latex'); ylim([0,140])
      set(gca,'TickLabelInterpreter','Latex','FontSize',fs)
      box off
  end

  for j = 1:3
    subplot(2,3,3+j), hold on
      stairs(t,X_data{i}(j,:)', '-'); 
      ylabel("$x_"+j+"(t)$ [cm]",'Interpreter','Latex'); ylim([0,60])
      xlabel('Time  - $t$ [s]','Interpreter','Latex')
      set(gca,'TickLabelInterpreter','Latex','FontSize',fs)
      box off
  end
end
subplot(2,3,6)
legend(["EE", "RK4", "CasADi", "Linear"], ...
        'Location','northwest','Interpreter','Latex', ...
        'FontSize',10, 'Box', 'off')
