function affine_model = Linearize(xdot, dT, xp, up, px)
% LINEARIZE Obtain a discrete-time linearized model from nonlinear dynamics
% 
%   AFFINE_MODEL = LINEARIZE(XDOT, dT, XP, UP, PX) obtains a DISCRETE-TIME
%       affine time-invariant model, AFFINE_MODEL = {c, A, B}, given by
%           x_k+1 = A x_k + B u_k + c
%        which corresponds to a (discretised) linear approximation of
%        the CONTINUOUS-TIME nonlinear dynamics with signature
%           xdot = f(t,x,u|px)
%                = XDOT(t,x,u,px)
%       around some specified fixed-point (xp,up).
%
%       Specifically, we compute the CONTINUOUS-TIME linearized model
%           xdot ≈ df/dx|_(xp,up)(x_k - xp) + df/du|_(xp,up)(u_k - up) + f(xp,up|px)
%                = A_tilde (x_k - xp) + B_tilde (u_k - up) + c_tilde
%       then convert this model to DISCRETE-TIME and re-arranges the constant
%       terms in the equation to avoid using deviation variables.
%

    import casadi.*

    Xp = MX.sym( ??? );     % <~ Define correctly the respective 
    Up = MX.sym( ??? );     % <~ symbols with their correct dimensions  

    f  = Function( ??? )    % <~ Fill with the function f(xp,up|px), using symbolic (Xp,Up)
    Jx = Function( ??? );   % <~ Fill with the function to compute Jacobian Jx = (df/dx)|_(xp,up)
    Ju = Function( ??? );   % <~ Fill with the function to compute Jacobian Ju = (df/du)|_(xp,up)

    c_tilde = ???;          % <~ Evaluate the function f(.) and the Jacobians 
    A_tilde = ???;          %    Jx(.) and Ju(.) at the fixed-point used for
    B_tilde = ???;          %    the linearization, i.e., point (xp,up)

    % (PLEASE, DO NOT MODIFY!)
    % The continuous-time linearized model is converted to a discrete-time model below
    %  (it can be done analytically because the model is linear/affine). Moreover, the
    %  code below also re-arranges all constant terms into the affine term (c_tilde), 
    %  so to avoid expressing the model in terms of deviation variables (x-xp).
    SS_dt = expm([A_tilde  B_tilde  c_tilde-A_tilde*xp-B_tilde*up; zeros(3,6)]*dT);
    A = SS_dt(1:3,1:3); 
    B = SS_dt(1:3,4:5);
    c = SS_dt(1:3,end);
    % --

    affine_model = {c, A, B};
% --
end